from abc import abstractmethod
import ast
from optparse import Option
from typing import List, Optional

from pydantic import BaseSettings
from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from turms.registry import ClassRegistry


class PluginConfig(BaseSettings):
    type: Optional[str]


class Plugin:
    """Base Plugin Class

    Raises:
        NotImplementedError: [description]
    """

    @abstractmethod
    def generate_ast(
        self,
        config: GeneratorConfig,
        client_schema: GraphQLSchema,
        registry: ClassRegistry,
    ) -> List[ast.AST]:
        raise NotImplementedError("Plugin must overrwrite this")

    def __str__(self) -> str:
        return self.__class__.__name__



from contextlib import contextmanager


try:
    from rich.console import Console
except ImportError:

    class Console:
        def __init__(self) -> None:
            pass

        def print(self, *args, **kwargs) -> None:
            print(*args, **kwargs)

        def warn(self, *args, **kwargs) -> None:
            print(*args, **kwargs)

        def status(self):
            @contextmanager
            def x(*args, **kwargs):
                yield x

            return x



from pydantic import BaseModel
from rich import get_console
from turms.processor.base import Processor


class BlackProcessorConfig(BaseModel):
    pass


class BlackProcessor(Processor):
    def __init__(self, config=None, **data):

        self.plugin_config = config or BlackProcessorConfig(**data)

    def run(self, gen_file: str):
        from black import format_str, FileMode

        return format_str(gen_file, mode=FileMode())



import ast
from distutils.command.config import config
from typing import List
from turms.config import GeneratorConfig
from keyword import iskeyword
from turms.errors import NoScalarEquivalentFound
from turms.stylers.base import Styler
from rich import get_console


SCALAR_DEFAULTS = {
    "ID": "str",
    "String": "str",
    "Int": "int",
    "Boolean": "bool",
    "GenericScalar": "typing.Dict",
    "DateTime": "str",
    "Float": "float",
}


class ClassRegistry(object):
    def __init__(self, config: GeneratorConfig, stylers: List[Styler]):
        self.stylers = stylers
        self._imports = set()
        self.config = config
        self.fragment_document_map = {}
        self.enum_class_map = {}
        self.inputtype_class_map = {}
        self.fragment_class_map = {}
        self.console = get_console()

    def generate_type_name_field(self, typename):
        return ast.AnnAssign(
            target=ast.Name(id="typename", ctx=ast.Store()),
            annotation=ast.Subscript(
                value=ast.Name(id="Optional", ctx=ast.Load()),
                slice=ast.Name("str", ctx=ast.Load()),
            ),
            value=ast.Call(
                func=ast.Name(id="Field", ctx=ast.Load()),
                args=[],
                keywords=[
                    ast.keyword(arg="alias", value=ast.Constant(value="__typename"))
                ],
            ),
            simple=1,
        )

    def generate_inputtype_classname(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_input_name(typename)

        if iskeyword(typename):
            return typename + "_"

        return typename

    def generate_node_name(self, node_name: str):
        for styler in self.stylers:
            node_name = styler.style_node_name(node_name)

        if iskeyword(node_name):
            return node_name + "_"

        return node_name

    def generate_parameter_name(self, node_name: str):
        for styler in self.stylers:
            node_name = styler.style_parameter_name(node_name)

        if iskeyword(node_name):
            return node_name + "_"

        return node_name

    def generate_fragment_classname(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_fragment_name(typename)

        if iskeyword(typename):
            return typename + "_"

        return typename

    def generate_enum_classname(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_enum_name(typename)

        if iskeyword(typename):
            return typename + "_"

        return typename

    def generate_query_classname(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_query_name(typename)

        if iskeyword(typename):
            return typename + "_"

        return typename

    def generate_mutation_classname(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_mutation_name(typename)

        if iskeyword(typename):
            return typename + "_"

        return typename

    def generate_subscription_classname(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_subscription_name(typename)

        if iskeyword(typename):
            return typename + "_"

        return typename

    def register_import(self, name):
        if name in ("bool", "str", "int", "float", "dict", "list", "tuple"):
            return

        assert "." in name, "Please only register imports with a top level package"
        self._imports.add(name)

    def generate_imports(self):
        imports = []

        top_level = {}
        for name in self._imports:
            top_level.setdefault(".".join(name.split(".")[:-1]), set()).add(
                name.split(".")[-1]
            )

        for top_level_name, sub_level_names in top_level.items():
            imports.append(
                ast.ImportFrom(
                    module=top_level_name,
                    names=[ast.alias(name=name) for name in sub_level_names],
                    level=0,
                )
            )

        return imports

    def register_enum_class(self, typename: str, cls: str):
        assert cls not in self.enum_class_map, f"{cls} already registered"
        self.enum_class_map[typename] = cls

    def register_inputtype_class(self, typename: str, cls: str):
        assert cls not in self.inputtype_class_map, f"{cls} already registered"
        self.inputtype_class_map[typename] = cls

    def register_fragment_class(self, typename: str, cls: str):
        assert cls not in self.fragment_class_map, f"{cls} already registered"
        self.fragment_class_map[typename] = cls

    def register_fragment_document(self, typename: str, cls: str):
        assert cls not in self.fragment_document_map, f"{cls} already registered"
        self.fragment_document_map[typename] = cls

    def get_inputtype_class(self, typename: str, allow_forward=True):
        try:
            return self.inputtype_class_map[typename]
        except KeyError as e:
            if allow_forward:
                return self.generate_inputtype_classname(typename)
            raise e

    def get_fragment_class(self, typename: str, allow_forward=True):
        try:
            return self.fragment_class_map[typename]
        except KeyError as e:
            if allow_forward:
                return self.generate_fragment_classname(typename)
            raise e

    def get_fragment_document(self, typename: str):
        return self.fragment_document_map[typename]

    def get_enum_class(self, typename: str, allow_forward=True):
        try:
            return self.enum_class_map[typename]
        except KeyError as e:
            if allow_forward:
                return self.generate_enum_classname(typename)
            raise e

    def get_scalar_equivalent(self, scalar_type: str):

        updated_dict = {**SCALAR_DEFAULTS, **self.config.scalar_definitions}

        try:
            scalar_type = updated_dict[scalar_type]
            if "." in scalar_type:
                # We make the assumption that the scalar type is a fully qualified class name
                self.register_import(scalar_type)
        except KeyError as e:
            raise NoScalarEquivalentFound(
                f"No python equivalent found for {scalar_type}. Please define in scalar_definitions"
            )

        return scalar_type.split(".")[-1]

    def warn(self, message):
        self.console.print("[r]" + message)



from turms.stylers.base import BaseStyler
import re


def camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


class SnakeNodeName(BaseStyler):
    def style_node_name(self, name: str) -> str:
        return camel_to_snake(name)

    def style_parameter_name(self, name: str) -> str:
        return camel_to_snake(name)



from pydantic import AnyHttpUrl, BaseModel, BaseSettings, Field
from typing import Dict, List, Optional, Union


class GeneratorConfig(BaseSettings):
    domain: str = "default"
    out_dir: str = "api"
    generated_name: str = "schema.py"
    documents: Optional[str]

    object_bases: List[str] = ["pydantic.BaseModel"]
    interface_bases: Optional[List[str]] = None
    always_resolve_interfaces: bool = True

    scalar_definitions = {}
    freeze: bool = False
    additional_bases = {}
    extensions: Dict = {}

    class Config:
        env_prefix = "TURMS_"
        extra = "allow"


class GraphQLConfig(BaseSettings):
    schema_url: Optional[Union[AnyHttpUrl, str]] = Field(alias="schema")
    bearer_token: Optional[str] = None
    documents: Optional[str]
    domain: str = "default"

    class Config:
        env_prefix = "TURMS_GRAPHQL_"
        extra = "allow"



from __future__ import annotations
from abc import abstractmethod
import ast
from typing import Any, List, Literal
from attr import Attribute

from graphql import (
    BooleanValueNode,
    FloatValueNode,
    FragmentSpreadNode,
    IntValueNode,
    NamedTypeNode,
    NullValueNode,
    StringValueNode,
    ValueNode,
    VariableDefinitionNode,
)
from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from turms.plugins.base import Plugin, PluginConfig
from pydantic import BaseModel, BaseSettings
from graphql.language.ast import OperationDefinitionNode, OperationType
from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from turms.plugins.base import Plugin
from pydantic import BaseModel
from graphql.language.ast import (
    FieldNode,
    ListTypeNode,
    NonNullTypeNode,
    OperationDefinitionNode,
    OperationType,
)
from turms.registry import ClassRegistry
from turms.utils import (
    NoDocumentsFoundError,
    generate_typename_field,
    parse_documents,
    target_from_node,
)
from turms.errors import NoScalarEquivalentFound
import re
import ast

from graphql.type.definition import (
    GraphQLList,
    get_named_type,
    is_list_type,
)
from graphql.utilities.get_operation_root_type import get_operation_root_type
from graphql.utilities.type_info import TypeInfo, get_field_def
import logging
from typing import Optional, Union


logger = logging.getLogger(__name__)


class Kwarg(BaseModel):
    key: str
    type: str
    description: str = "Specify that in turms.plugin.funcs.OperationsFuncPlugin"
    default: Any = None


class Arg(BaseModel):
    key: str
    type: str
    description: str = "Specify that in turms.plugin.funcs.OperationsFuncPlugin"


class FunctionDefinition(BaseModel):
    type: OperationType
    is_async: bool = False
    extra_args: List[Arg] = []
    extra_kwargs: List[Kwarg] = []
    use: str


class OperationsFuncPluginConfig(PluginConfig):
    funcs_glob: Optional[str]
    prepend_sync: str = ""
    prepend_async: str = "a"
    collapse_lonely: bool = True
    global_args: List[Arg] = []
    global_kwargs: List[Kwarg] = []
    definitions: List[FunctionDefinition] = []

    class Config:
        env_prefix = "TURMS_PLUGINS_FUNCS_"


def camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def generate_async_func_name(
    o: OperationDefinitionNode,
    plugin_config: OperationsFuncPluginConfig,
    config: GeneratorConfig,
    registry: ClassRegistry,
):

    return f"{plugin_config.prepend_async}{camel_to_snake(o.name.value)}"


def generate_sync_func_name(
    o: OperationDefinitionNode,
    plugin_config: OperationsFuncPluginConfig,
    config: GeneratorConfig,
    registry: ClassRegistry,
):

    return f"{plugin_config.prepend_sync}{camel_to_snake(o.name.value)}"


def get_input_type_annotation(
    input_type: NamedTypeNode,
    config: GeneratorConfig,
    registry: ClassRegistry,
    optional=True,
):

    if isinstance(input_type, NamedTypeNode):
        try:
            type_name = registry.get_scalar_equivalent(input_type.name.value)
        except NoScalarEquivalentFound as e:
            type_name = registry.get_inputtype_class(input_type.name.value)

        if optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name(id="Optional", ctx=ast.Load()),
                slice=ast.Name(
                    id=type_name,
                    ctx=ast.Load(),
                ),
            )

        return ast.Name(
            id=type_name,
            ctx=ast.Load(),
        )

    elif isinstance(input_type, ListTypeNode):
        registry.register_import("typing.List")

        if optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name(id="Optional", ctx=ast.Load()),
                slice=ast.Subscript(
                    value=ast.Name(id="List", ctx=ast.Load()),
                    slice=get_input_type_annotation(input_type.type, config, registry),
                ),
            )
        return ast.Subscript(
            value=ast.Name(id="List", ctx=ast.Load()),
            slice=get_input_type_annotation(input_type.type, config, registry),
        )

    elif isinstance(input_type, NonNullTypeNode):
        return get_input_type_annotation(
            input_type.type, config, registry, optional=False
        )

    raise NotImplementedError()


wanted_definition = {
    OperationType.MUTATION: {
        "async": "asyncfunction",
        "sync": "function",
    },
    OperationType.QUERY: {
        "async": "asyncfunction",
        "sync": "function",
    },
    OperationType.SUBSCRIPTION: {
        "async": "asynciterator",
        "sync": "iterator",
    },
}


def get_extra_args_for_onode(
    definition: FunctionDefinition,
    plugin_config: OperationsFuncPluginConfig,
) -> List[Arg]:

    args = plugin_config.global_args
    return args + definition.extra_args


def generate_passing_extra_args_for_onode(
    definition: FunctionDefinition, plugin_config: OperationsFuncPluginConfig
):

    return [
        ast.Name(id=arg.key, ctx=ast.Load())
        for arg in get_extra_args_for_onode(definition, plugin_config)
    ]


def generate_passing_extra_kwargs_for_onode(
    definition: FunctionDefinition, plugin_config: OperationsFuncPluginConfig
):

    return [
        ast.keyword(arg=kwarg.key, value=ast.Name(id=kwarg.key, ctx=ast.Load()))
        for kwarg in get_extra_kwargs_for_onode(definition, plugin_config)
    ]


def get_extra_kwargs_for_onode(
    definition: FunctionDefinition,
    plugin_config: OperationsFuncPluginConfig,
) -> List[Arg]:

    kwargs = plugin_config.global_kwargs

    return kwargs + definition.extra_kwargs


def parse_value_node(x: ValueNode):
    if isinstance(x, IntValueNode):
        return int(x.value)
    elif isinstance(x, FloatValueNode):
        return float(x.value)
    elif isinstance(x, StringValueNode):
        return x.value
    elif isinstance(x, BooleanValueNode):
        return x.value == "true"
    elif isinstance(x, NullValueNode):
        return None
    else:
        raise NotImplementedError(f"cannot parse {x}")


def get_definitions_for_onode(
    o: OperationDefinitionNode,
    plugin_config: OperationsFuncPluginConfig,
) -> List[Arg]:

    definitions = [
        definition
        for definition in plugin_config.definitions
        if definition.type == o.operation
    ]

    return definitions


def generate_query_args(
    definition: FunctionDefinition,
    o: OperationDefinitionNode,
    config: GeneratorConfig,
    plugin_config: OperationsFuncPluginConfig,
    registry: ClassRegistry,
):

    extra_args = get_extra_args_for_onode(definition, plugin_config)
    pos_args = []

    for arg in extra_args:
        registry.register_import(arg.type)
        pos_args.append(
            ast.arg(
                arg=arg.key,
                annotation=ast.Name(
                    id=arg.type.split(".")[-1],
                    ctx=ast.Load(),
                ),
            )
        )

    for v in o.variable_definitions:
        if isinstance(v.type, NonNullTypeNode) and not v.default_value:
            pos_args.append(
                ast.arg(
                    arg=registry.generate_parameter_name(v.variable.name.value),
                    annotation=get_input_type_annotation(v.type.type, config, registry),
                )
            )

    kw_args = []
    kw_values = []

    for v in o.variable_definitions:
        if not isinstance(v.type, NonNullTypeNode) or v.default_value:
            kw_args.append(
                ast.arg(
                    arg=registry.generate_parameter_name(v.variable.name.value),
                    annotation=get_input_type_annotation(v.type, config, registry),
                )
            )
            kw_values.append(
                ast.Constant(
                    value=parse_value_node(v.default_value) if v.default_value else None
                )
            )

    extra_kwargs = get_extra_kwargs_for_onode(definition, plugin_config)

    for kwarg in extra_kwargs:
        registry.register_import(kwarg.type)
        kw_args.append(
            ast.arg(
                arg=kwarg.key,
                annotation=ast.Name(
                    id=kwarg.type.split(".")[-1],
                    ctx=ast.Load(),
                ),
            )
        )
        kw_values.append(ast.Constant(value=kwarg.default))

    return ast.arguments(
        args=pos_args + kw_args,
        posonlyargs=[],
        kwonlyargs=[],
        kw_defaults=[],
        defaults=kw_values,
    )


def generate_query_dict(o: OperationDefinitionNode, registry: ClassRegistry):

    keys = []
    values = []

    for v in o.variable_definitions:
        keys.append(ast.Constant(value=v.variable.name.value))
        values.append(
            ast.Name(
                id=registry.generate_parameter_name(v.variable.name.value),
                ctx=ast.Load(),
            )
        )

    return ast.Dict(keys=keys, values=values)


def generate_document_arg(o_name):

    return ast.Name(id=o_name, ctx=ast.Load())


def recurse_variable_annotation(
    v: VariableDefinitionNode, registry: ClassRegistry, optional=True
):

    if isinstance(v.type, NamedTypeNode):
        try:
            x = registry.get_scalar_equivalent(v.type.name.value)
        except NoScalarEquivalentFound as e:
            x = registry.get_inputtype_class(v.type.name.value)
        if optional:
            return "Optional[" + x + "]"
        else:
            return x

    elif isinstance(v.type, NonNullTypeNode):
        return recurse_variable_annotation(v.type, registry, optional=False)

    elif isinstance(v.type, ListTypeNode):
        if optional:
            return (
                "Optional[List[" + recurse_variable_annotation(v.type, registry) + "]]"
            )
        return "List[" + recurse_variable_annotation(v.type, registry) + "]"

    raise NotImplementedError()


def generate_query_doc(
    definition: FunctionDefinition,
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: OperationsFuncPluginConfig,
    registry: ClassRegistry,
    collapse=False,
):

    x = get_operation_root_type(client_schema, o)
    o.__annotations__

    if o.operation == OperationType.QUERY:
        o_name = registry.generate_query_classname(o.name.value)
    if o.operation == OperationType.MUTATION:
        o_name = registry.generate_mutation_classname(o.name.value)
    if o.operation == OperationType.SUBSCRIPTION:
        o_name = registry.generate_subscription_classname(o.name.value)

    if collapse:
        # Check if was collapsed from fragment

        return_type = f"{o_name}{o.selection_set.selections[0].name.value.capitalize()}"
        field_definition = get_field_def(
            client_schema, x, o.selection_set.selections[0]
        )

        potential_item = o.selection_set.selections[0]

        if potential_item.selection_set is None:
            return_type = registry.get_scalar_equivalent(field_definition.type.name)

        elif len(potential_item.selection_set.selections) == 1:
            if isinstance(
                potential_item.selection_set.selections[0], FragmentSpreadNode
            ):
                return_type = registry.get_fragment_class(
                    potential_item.selection_set.selections[0].name.value
                )

    else:

        return_type = f"{o_name}"

    header = f"{o.name.value} \n\n"

    if collapse:
        field = o.selection_set.selections[0]
        field_definition = get_field_def(client_schema, x, field)
        description = (
            header + field_definition.description
            if field_definition.description
            else header
        )

    else:
        query_descriptions = []

        for field in o.selection_set.selections:
            if isinstance(field, FieldNode):
                target = target_from_node(field)
                field_definition = get_field_def(client_schema, x, field)
                if field_definition.description:
                    query_descriptions.append(
                        f"{target}: {field_definition.description}"
                    )

        description = "\n ".join([header] + query_descriptions)

    description += "\n\nArguments:\n"

    extra_args = get_extra_args_for_onode(definition, plugin_config)

    for arg in extra_args:
        description += f"    {arg.key} ({arg.type}): {arg.description}\n"

    for v in o.variable_definitions:
        if isinstance(v.type, NonNullTypeNode) and not v.default_value:
            description += f"    {registry.generate_parameter_name(v.variable.name.value)} ({recurse_variable_annotation(v, registry)}): {v.variable.name.value}\n"

    for v in o.variable_definitions:
        if not isinstance(v.type, NonNullTypeNode) or v.default_value:
            description += f"    {registry.generate_parameter_name(v.variable.name.value)} ({recurse_variable_annotation(v, registry)}, optional): {v.variable.name.value}. {'' if not v.default_value else  'Defaults to ' + str(v.default_value.value)}\n"

    extra_kwargs = get_extra_kwargs_for_onode(definition, plugin_config)
    for kwarg in extra_kwargs:
        description += (
            f"    {kwarg.key} ({kwarg.type}, optional): {kwarg.description}\n"
        )

    description += "\nReturns:\n"
    description += f"    {return_type}"

    return ast.Expr(value=ast.Constant(value=description))


def genereate_async_call(
    definition: FunctionDefinition,
    o_name: str,
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: OperationsFuncPluginConfig,
    registry: ClassRegistry,
    collapse=False,
):
    registry.register_import(definition.use)
    if not collapse:
        return ast.Return(
            value=ast.Await(
                value=ast.Call(
                    func=ast.Name(
                        id=definition.use.split(".")[-1],
                        ctx=ast.Load(),
                    ),
                    keywords=generate_passing_extra_kwargs_for_onode(
                        definition, plugin_config
                    ),
                    args=generate_passing_extra_args_for_onode(
                        definition, plugin_config
                    )
                    + [
                        generate_document_arg(o_name),
                        generate_query_dict(o, registry),
                    ],
                )
            )
        )
    else:

        return ast.Return(
            value=ast.Attribute(
                value=ast.Await(
                    value=ast.Call(
                        func=ast.Name(
                            id=definition.use.split(".")[-1],
                            ctx=ast.Load(),
                        ),
                        keywords=generate_passing_extra_kwargs_for_onode(
                            definition, plugin_config
                        ),
                        args=generate_passing_extra_args_for_onode(
                            definition, plugin_config
                        )
                        + [
                            generate_document_arg(o_name),
                            generate_query_dict(o, registry),
                        ],
                    )
                ),
                attr=registry.generate_node_name(
                    o.selection_set.selections[0].name.value
                ),
                ctx=ast.Load(),
            )
        )


def genereate_sync_call(
    definition: FunctionDefinition,
    o_name: str,
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: OperationsFuncPluginConfig,
    registry: ClassRegistry,
    collapse=False,
):
    registry.register_import(definition.use)
    if not collapse:
        return ast.Return(
            value=ast.Call(
                func=ast.Name(
                    id=definition.use.split(".")[-1],
                    ctx=ast.Load(),
                ),
                keywords=generate_passing_extra_kwargs_for_onode(
                    definition, plugin_config
                ),
                args=generate_passing_extra_args_for_onode(definition, plugin_config)
                + [
                    generate_document_arg(o_name),
                    generate_query_dict(o, registry),
                ],
            )
        )
    else:
        return ast.Return(
            value=ast.Attribute(
                value=ast.Call(
                    func=ast.Name(
                        id=definition.use.split(".")[-1],
                        ctx=ast.Load(),
                    ),
                    keywords=generate_passing_extra_kwargs_for_onode(
                        definition, plugin_config
                    ),
                    args=generate_passing_extra_args_for_onode(
                        definition, plugin_config
                    )
                    + [
                        generate_document_arg(o_name),
                        generate_query_dict(o, registry),
                    ],
                ),
                attr=registry.generate_node_name(
                    o.selection_set.selections[0].name.value
                ),
                ctx=ast.Load(),
            )
        )


def genereate_async_iterator(
    definition: FunctionDefinition,
    o_name: str,
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: OperationsFuncPluginConfig,
    registry: ClassRegistry,
    collapse=False,
):
    registry.register_import(definition.use)
    if not collapse:
        return ast.AsyncFor(
            target=ast.Name(id="event", ctx=ast.Store()),
            iter=ast.Call(
                func=ast.Name(
                    id=definition.use.split(".")[-1],
                    ctx=ast.Load(),
                ),
                keywords=generate_passing_extra_kwargs_for_onode(
                    definition, plugin_config
                ),
                args=generate_passing_extra_args_for_onode(definition, plugin_config)
                + [
                    generate_document_arg(o_name),
                    generate_query_dict(o, registry),
                ],
            ),
            body=[
                ast.Expr(value=ast.Yield(value=ast.Name(id="event", ctx=ast.Load()))),
            ],
            orelse=[],
        )
    else:
        return ast.AsyncFor(
            target=ast.Name(id="event", ctx=ast.Store()),
            iter=ast.Call(
                func=ast.Name(
                    id=definition.use.split(".")[-1],
                    ctx=ast.Load(),
                ),
                keywords=generate_passing_extra_kwargs_for_onode(
                    definition, plugin_config
                ),
                args=generate_passing_extra_args_for_onode(definition, plugin_config)
                + [
                    generate_document_arg(o_name),
                    generate_query_dict(o, registry),
                ],
            ),
            body=[
                ast.Expr(
                    value=ast.Yield(
                        value=ast.Attribute(
                            value=ast.Name(id="event", ctx=ast.Load()),
                            ctx=ast.Load(),
                            attr=registry.generate_node_name(
                                o.selection_set.selections[0].name.value
                            ),
                        )
                    )
                ),
            ],
            orelse=[],
        )


def genereate_sync_iterator(
    definition: FunctionDefinition,
    o_name: str,
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: OperationsFuncPluginConfig,
    registry: ClassRegistry,
    collapse=False,
):
    registry.register_import(definition.use)
    if not collapse:
        return ast.For(
            target=ast.Name(id="event", ctx=ast.Store()),
            iter=ast.Call(
                func=ast.Name(
                    id=definition.use.split(".")[-1],
                    ctx=ast.Load(),
                ),
                keywords=generate_passing_extra_kwargs_for_onode(
                    definition, plugin_config
                ),
                args=generate_passing_extra_args_for_onode(definition, plugin_config)
                + [
                    generate_document_arg(o_name),
                    generate_query_dict(o, registry),
                ],
            ),
            body=[
                ast.Expr(value=ast.Yield(value=ast.Name(id="event", ctx=ast.Load()))),
            ],
            orelse=[],
        )
    else:
        return ast.For(
            target=ast.Name(id="event", ctx=ast.Store()),
            iter=ast.Call(
                func=ast.Name(
                    id=definition.use.split(".")[-1],
                    ctx=ast.Load(),
                ),
                keywords=generate_passing_extra_kwargs_for_onode(
                    definition, plugin_config
                ),
                args=generate_passing_extra_args_for_onode(definition, plugin_config)
                + [
                    generate_document_arg(o_name),
                    generate_query_dict(o, registry),
                ],
            ),
            body=[
                ast.Expr(
                    value=ast.Yield(
                        value=ast.Attribute(
                            value=ast.Name(id="event", ctx=ast.Load()),
                            ctx=ast.Load(),
                            attr=registry.generate_node_name(
                                o.selection_set.selections[0].name.value
                            ),
                        )
                    )
                ),
            ],
            orelse=[],
        )


def generate_operation_func(
    definition: FunctionDefinition,
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: OperationsFuncPluginConfig,
    registry: ClassRegistry,
):
    tree = []

    if o.operation == OperationType.QUERY:
        o_name = registry.generate_query_classname(o.name.value)
    if o.operation == OperationType.MUTATION:
        o_name = registry.generate_mutation_classname(o.name.value)
    if o.operation == OperationType.SUBSCRIPTION:
        o_name = registry.generate_subscription_classname(o.name.value)

    collapse = len(o.selection_set.selections) == 1 and plugin_config.collapse_lonely

    return_type = (
        f"{o_name}{o.selection_set.selections[0].name.value.capitalize()}"
        if collapse
        else o_name
    )

    if collapse:
        x = get_operation_root_type(client_schema, o)
        field_definition = get_field_def(
            client_schema, x, o.selection_set.selections[0]
        )

        potential_item = o.selection_set.selections[0]

        # Check if was collapsed from fragment
        if potential_item.selection_set is None:
            return_type = registry.get_scalar_equivalent(field_definition.type.name)

        elif len(potential_item.selection_set.selections) == 1:
            if isinstance(
                potential_item.selection_set.selections[0], FragmentSpreadNode
            ):
                return_type = registry.get_fragment_class(
                    potential_item.selection_set.selections[0].name.value
                )
        else:
            return_type = (
                f"{o_name}{o.selection_set.selections[0].name.value.capitalize()}"
            )

        if isinstance(field_definition.type, GraphQLList):
            returns = ast.Subscript(
                value=ast.Name(id="List", ctx=ast.Load()),
                slice=ast.Name(id=return_type, ctx=ast.Load()),
            )
        else:
            returns = ast.Name(id=return_type, ctx=ast.Load())
    else:
        returns = ast.Name(id=return_type, ctx=ast.Load())

    doc = generate_query_doc(
        definition, o, client_schema, config, plugin_config, registry, collapse
    )

    if definition.is_async:
        if o.operation == OperationType.SUBSCRIPTION:
            registry.register_import("typing.AsyncIterator")

        tree.append(
            ast.AsyncFunctionDef(
                name=generate_async_func_name(o, plugin_config, config, registry),
                args=generate_query_args(
                    definition,
                    o,
                    config,
                    plugin_config,
                    registry,
                ),
                body=[
                    doc,
                    genereate_async_call(
                        definition,
                        o_name,
                        o,
                        client_schema,
                        config,
                        plugin_config,
                        registry,
                        collapse,
                    )
                    if definition.type != OperationType.SUBSCRIPTION
                    else genereate_async_iterator(
                        definition,
                        o_name,
                        o,
                        client_schema,
                        config,
                        plugin_config,
                        registry,
                        collapse,
                    ),
                ],
                decorator_list=[],
                returns=returns
                if definition.type != OperationType.SUBSCRIPTION
                else ast.Subscript(
                    value=ast.Name(id="AsyncIterator", ctx=ast.Load()), slice=returns
                ),
            )
        )

    if not definition.is_async:
        if o.operation == OperationType.SUBSCRIPTION:
            registry.register_import("typing.Iterator")

        tree.append(
            ast.FunctionDef(
                name=generate_sync_func_name(o, plugin_config, config, registry),
                args=generate_query_args(
                    definition,
                    o,
                    config,
                    plugin_config,
                    registry,
                ),
                body=[
                    doc,
                    genereate_sync_call(
                        definition,
                        o_name,
                        o,
                        client_schema,
                        config,
                        plugin_config,
                        registry,
                        collapse,
                    )
                    if definition.type != OperationType.SUBSCRIPTION
                    else genereate_sync_iterator(
                        definition,
                        o_name,
                        o,
                        client_schema,
                        config,
                        plugin_config,
                        registry,
                        collapse,
                    ),
                ],
                decorator_list=[],
                returns=returns
                if definition.type != OperationType.SUBSCRIPTION
                else ast.Subscript(
                    value=ast.Name(id="Iterator", ctx=ast.Load()), slice=returns
                ),
            )
        )

    return tree


class OperationsFuncPlugin(Plugin):
    def __init__(self, config=None, **data):
        self.plugin_config = config or OperationsFuncPluginConfig(**data)

    def generate_ast(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        registry: ClassRegistry,
    ) -> List[ast.AST]:

        plugin_tree = []

        try:
            documents = parse_documents(
                client_schema, self.plugin_config.funcs_glob or config.documents
            )
        except NoDocumentsFoundError as e:
            logger.exception(e)
            return plugin_tree

        operations = [
            node
            for node in documents.definitions
            if isinstance(node, OperationDefinitionNode)
        ]

        for operation in operations:
            for definition in get_definitions_for_onode(operation, self.plugin_config):
                plugin_tree += generate_operation_func(
                    definition,
                    operation,
                    client_schema,
                    config,
                    self.plugin_config,
                    registry,
                )

        return plugin_tree



import asyncio
import asyncio
import sys
import time
from rich.console import Console
from watchdog.observers import Observer
from watchdog.events import (
    FileModifiedEvent,
    FileSystemEventHandler,
)
import os
import subprocess
import janus
import threading
import os
import signal
import yaml

from turms.config import GraphQLConfig
from turms.run import gen


console = Console()


class QueueHandler(FileSystemEventHandler):
    def __init__(self, *args, sync_q=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.queue = sync_q

    def on_any_event(self, event):
        self.queue.put(event)
        self.queue.join()


def watcher(path, queue, event: threading.Event):
    try:
        console.print(f"Watching path {path}")
        event_handler = QueueHandler(sync_q=queue)
        observer = Observer()
        observer.schedule(event_handler, path, recursive=True)
        observer.start()

        with console.status("Watching"):
            while not event.is_set():
                time.sleep(1)
    except:
        print("Watcher failed")


async def buffered_queue(queue, timeout=4):
    buffer = [None]

    async def iterator():
        while True:
            nana = await queue.get()
            queue.task_done()
            if isinstance(nana, FileModifiedEvent):
                console.print("modified file")
                if nana.src_path.lower().endswith((".graphql")):
                    buffer.append(nana)

    loop = asyncio.get_event_loop()
    loop.create_task(iterator())
    while True:
        await asyncio.sleep(timeout)
        yield buffer[len(buffer) - 1]
        buffer = [None]


class Host:
    def __init__(self, path=None, config_path=None, project=None) -> None:
        self.config_path = config_path
        self.watch_path = os.path.join(
            os.getcwd(),
            "/".join(
                path.split("*")[0].split("/")[:-1]
            ),  # should get root path even if using glob
        )
        self.console = Console()
        self.project = project

    async def restart(self):
        console.print("Restarting Generation")
        try:
            gen(self.config_path, project=self.project)
        except Exception as e:
            console.print("Generation Failed")
            console.print_exception()

    async def run(self):
        jqueue = janus.Queue()

        cancel_event = threading.Event()
        loop = asyncio.get_running_loop()
        fut = loop.run_in_executor(
            None, watcher, self.watch_path, jqueue.sync_q, cancel_event
        )

        try:
            async for event in buffered_queue(jqueue.async_q, timeout=1):
                if event:
                    await self.restart()
        except asyncio.CancelledError as e:
            cancel_event.set()
            await fut

        jqueue.close()
        await jqueue.wait_closed()


def watch(filepath, project=None):

    with open(filepath, "r") as f:
        yaml_dict = yaml.safe_load(f)

    assert "projects" in yaml_dict, "Right now only projects is supported"

    project = project or list(yaml_dict["projects"].items())[0][0]

    config = GraphQLConfig(**yaml_dict["projects"][project], domain=project)
    console.print(f"Running filewatcher for project: {project}")

    host = Host(path=config.documents, config_path=filepath, project=project)
    asyncio.run(host.run())



from graphql import GraphQLSchema, get_introspection_query, parse, build_ast_schema
from graphql.language.parser import parse
from pydantic import AnyHttpUrl, NoneIsNotAllowedError
from rich import get_console
from turms.config import GeneratorConfig, GraphQLConfig
from turms.helpers import (
    build_schema_from_glob,
    build_schema_from_introspect_url,
    import_string,
)
from turms.processor.base import Processor
from turms.processor.black import BlackProcessor
from turms.registry import ClassRegistry
from turms.stylers import Styler
from turms.stylers.default import DefaultStyler
from turms.utils import build_schema
from turms.plugins.base import Plugin
from turms.compat.funcs import unparse
from typing import Dict, List
import ast
import yaml
import json
import os
from urllib import request
from urllib.error import URLError
import glob
from .errors import GenerationError

plugins: List[Plugin] = []

processors: List[Processor] = []
stylers: List[Styler] = []


def gen(filepath: str, project=None):

    with open(filepath, "r") as f:
        yaml_dict = yaml.safe_load(f)

    assert "projects" in yaml_dict, "Right now only projects is supported"

    for key, project in yaml_dict["projects"].items():
        try:
            get_console().print(
                f"-------------- Generating project: {key} --------------"
            )
            config = GraphQLConfig(**project, domain=key)

            if isinstance(config.schema_url, AnyHttpUrl):
                schema = build_schema_from_introspect_url(
                    config.schema_url, config.bearer_token
                )
            else:
                schema = build_schema_from_glob(config.schema_url)

            turms_config = project["extensions"]["turms"]

            gen_config = GeneratorConfig(
                **turms_config, documents=project["documents"], domain=config.domain
            )

            plugins = []
            stylers = []
            processors = []

            if "plugins" in turms_config:
                plugin_configs = turms_config["plugins"]
                for plugin_config in plugin_configs:
                    assert (
                        "type" in plugin_config
                    ), "A plugin must at least specify type"
                    plugin_class = import_string(plugin_config["type"])
                    plugins.append(plugin_class(**plugin_config))
                    get_console().print(f"Using Plugin {plugin_class}")

            if "stylers" in turms_config:
                plugin_configs = turms_config["stylers"]
                for plugin_config in plugin_configs:
                    assert (
                        "type" in plugin_config
                    ), "A styler must at least specify type"
                    styler_class = import_string(plugin_config["type"])
                    stylers.append(styler_class(**plugin_config))
                    get_console().print(f"Using Styler {styler_class}")

            else:
                stylers.append(DefaultStyler())

            if "processors" in turms_config:
                proc_configs = turms_config["processors"]
                for proc_config in proc_configs:
                    assert (
                        "type" in proc_config
                    ), "A processor must at least specify type"
                    proc_class = import_string(proc_config["type"])
                    processors.append(proc_class(**proc_config))
                    get_console().print(f"Using Processor {proc_class}")

            generated_ast = generate_ast(
                gen_config,
                schema,
                plugins=plugins,
                stylers=stylers,
            )

            md = ast.Module(body=generated_ast, type_ignores=[])
            generated = unparse(ast.fix_missing_locations(md))

            for processor in processors:
                generated = processor.run(generated)

            if not os.path.isdir(gen_config.out_dir):
                os.makedirs(gen_config.out_dir)

            with open(
                os.path.join(gen_config.out_dir, gen_config.generated_name), "w"
            ) as f:
                f.write(generated)

            get_console().print("Sucessfull!! :right-facing_fist::left-facing_fist:")
        except:
            get_console().print(
                "-------- ERROR FOR PROJECT: " + key.upper() + " --------"
            )
            get_console().print_exception()
            raise


def generate_ast(
    config: GeneratorConfig,
    schema: GraphQLSchema,
    plugins=plugins,
    stylers=stylers,
):

    global_tree = []
    registry = ClassRegistry(config, stylers)

    for plugin in plugins:
        try:
            global_tree += plugin.generate_ast(schema, config, registry)
        except Exception as e:
            raise GenerationError(f"Plugin:{plugin} failed!") from e

    global_tree = registry.generate_imports() + global_tree

    return global_tree



from abc import abstractmethod


class Styler:
    @abstractmethod
    def style_subscription_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")

    @abstractmethod
    def style_mutation_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")

    @abstractmethod
    def style_query_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")

    @abstractmethod
    def style_enum_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")

    @abstractmethod
    def style_fragment_class_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")

    @abstractmethod
    def style_node_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")

    @abstractmethod
    def style_parameter_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")

    @abstractmethod
    def style_input_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")


class BaseStyler(Styler):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

    def style_operation_name(self, name: str) -> str:
        return name

    def style_query_name(self, name: str) -> str:
        return name

    def style_mutation_name(self, name: str) -> str:
        return name

    def style_enum_name(self, name: str) -> str:
        return name

    def style_fragment_name(self, name: str) -> str:
        return name

    def style_node_name(self, name: str) -> str:
        return name

    def style_input_name(self, name: str) -> str:
        return name

    def style_parameter_name(self, name: str) -> str:
        return name

    def style_subscription_name(self, name: str) -> str:
        return name



from turms.stylers.base import BaseStyler
import re


def camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


class DefaultStyler(BaseStyler):
    def style_fragment_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_query_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_subscription_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_mutation_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_input_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_enum_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_node_name(self, name: str) -> str:
        return camel_to_snake(name)

    def style_parameter_name(self, name: str) -> str:
        return camel_to_snake(name)



import glob
import re
from typing import Dict, List
from turms.config import GeneratorConfig, GraphQLConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from graphql.language.ast import DocumentNode, FieldNode
from graphql.error.graphql_error import GraphQLError
from graphql import (
    language,
    parse,
    get_introspection_query,
    validate,
    build_client_schema,
)
import ast
from turms.registry import ClassRegistry


class FragmentNotFoundError(Exception):
    pass


class NoDocumentsFoundError(Exception):
    pass


class NoScalarEquivalentDefined(Exception):
    pass


def target_from_node(node: FieldNode) -> str:
    return (
        node.alias.value if hasattr(node, "alias") and node.alias else node.name.value
    )


def build_schema(introspection_query: Dict[str, str]):
    return build_client_schema(introspection_query)


def generate_typename_field(typename, registry: ClassRegistry):

    registry.register_import("pydantic.Field")
    registry.register_import("typing.Optional")
    registry.register_import("typing.Literal")

    return ast.AnnAssign(
        target=ast.Name(id="typename", ctx=ast.Store()),
        annotation=ast.Subscript(
            value=ast.Name(id="Optional", ctx=ast.Load()),
            slice=ast.Subscript(
                value=ast.Name("Literal", ctx=ast.Load()),
                slice=ast.Constant(value=typename),
            ),
        ),
        value=ast.Call(
            func=ast.Name(id="Field", ctx=ast.Load()),
            args=[],
            keywords=[ast.keyword(arg="alias", value=ast.Constant(value="__typename"))],
        ),
        simple=1,
    )


def generate_config_class(config: GeneratorConfig):

    config_fields = []

    if config.freeze:
        config_fields.append(
            ast.Assign(
                targets=[ast.Name(id="frozen", ctx=ast.Store())],
                value=ast.Constant(value=True),
            )
        )

    if len(config_fields) > 0:
        return [
            ast.ClassDef(
                name="Config",
                bases=[],
                keywords=[],
                body=config_fields,
                decorator_list=[],
            )
        ]
    else:
        return []


def parse_documents(
    client_schema: GraphQLSchema, scan_glob="g/*/**.graphql"
) -> DocumentNode:
    """ """

    x = glob.glob(scan_glob, recursive=True)

    errors: List[GraphQLError] = []

    dsl = ""

    for file in x:
        with open(file, "r") as f:
            dsl += f.read()

    if dsl == "":
        raise NoDocumentsFoundError(
            f"Glob {scan_glob} did not find documents. Or only empty documents"
        )

    nodes = parse(dsl)

    errors = validate(client_schema, nodes)
    if len(errors) > 0:
        raise Exception(errors)

    return nodes


fragment_searcher = re.compile(r"\.\.\.(?P<fragment>[a-zA-Z]*)")


def replace_iteratively(
    pattern,
    registry,
    taken=[],
):
    z = set(fragment_searcher.findall(pattern))  # only set is important
    new_fragments = [new_f for new_f in z if new_f not in taken and new_f != ""]
    if not new_fragments:
        return pattern
    else:
        try:
            level_down_pattern = "\n\n".join(
                [registry.get_fragment_document(key) for key in new_fragments]
                + [pattern]
            )
            return replace_iteratively(
                level_down_pattern, registry, taken=new_fragments + taken
            )
        except KeyError as e:
            raise FragmentNotFoundError(
                f"Could not find in Fragment Map {registry}"
            ) from e


def get_additional_bases_for_type(
    typename, config: GeneratorConfig, registry: ClassRegistry
):

    if typename in config.additional_bases:
        for base in config.additional_bases[typename]:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in config.additional_bases[typename]
        ]
    return []


def get_interface_bases(config: GeneratorConfig, registry: ClassRegistry):

    if config.interface_bases:
        for base in config.interface_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in config.interface_bases
        ]
    else:
        for base in config.object_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in config.object_bases
        ]



from pydantic import BaseModel
from rich import get_console
from turms.processor.base import Processor


class IsortProcessorConfig(BaseModel):
    pass


class IsortProcessor(Processor):
    def __init__(self, config=None, **data):
        self.plugin_config = config or IsortProcessorConfig(**data)

    def run(self, gen_file: str):
        import isort

        return isort.code(gen_file)



from pydantic import BaseModel
from turms.stylers.base import BaseStyler


class Capitalizer(BaseStyler):
    def __init__(self, **kwargs) -> None:
        super().__init__()

    def style_fragment_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_query_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_subscription_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_mutation_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_input_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_enum_name(self, typename):
        return typename[0].upper() + typename[1:]



from abc import abstractmethod
import ast
from typing import List, Optional
from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from graphql.language.ast import OperationDefinitionNode, OperationType
from turms.parser.recurse import recurse_annotation, type_field_node
from turms.plugins.base import Plugin, PluginConfig
from pydantic import BaseModel, BaseSettings
from graphql.error.graphql_error import GraphQLError
from graphql.error.syntax_error import GraphQLSyntaxError
from graphql.language.ast import (
    DefinitionNode,
    DocumentNode,
    FieldNode,
    FragmentDefinitionNode,
    FragmentSpreadNode,
    InlineFragmentNode,
    ListTypeNode,
    NonNullTypeNode,
    OperationDefinitionNode,
    OperationType,
    SelectionNode,
    SelectionSetNode,
)
from graphql.type.definition import (
    GraphQLEnumType,
    GraphQLField,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLType,
    get_named_type,
    is_list_type,
)
from graphql.type.validate import get_operation_type_node
from graphql.utilities.build_client_schema import build_client_schema, GraphQLSchema
from graphql.utilities.get_operation_root_type import get_operation_root_type
from graphql.utilities.type_info import TypeInfo, get_field_def

from pydantic.main import BaseModel
import re
from graphql import language, parse, get_introspection_query, validate
from turms.registry import ClassRegistry
from turms.utils import (
    NoDocumentsFoundError,
    generate_config_class,
    parse_documents,
    replace_iteratively,
)
import logging


logger = logging.getLogger(__name__)
fragment_searcher = re.compile(r"\.\.\.(?P<fragment>[a-zA-Z]*)")


class OperationsPluginConfig(PluginConfig):
    query_bases: List[str] = None
    mutation_bases: List[str] = None
    subscription_bases: List[str] = None
    operations_glob: Optional[str]

    class Config:
        env_prefix = "TURMS_PLUGINS_OPERATIONS_"


def get_query_bases(
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):

    if plugin_config.query_bases:
        for base in plugin_config.query_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in plugin_config.query_bases
        ]
    else:
        for base in config.object_bases:
            registry.register_import(base)

            return [
                ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                for base in config.object_bases
            ]


def get_mutation_bases(
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):

    if plugin_config.mutation_bases:
        for base in plugin_config.mutation_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in plugin_config.mutation_bases
        ]
    else:
        for base in config.object_bases:
            registry.register_import(base)

            return [
                ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                for base in config.object_bases
            ]


def get_subscription_bases(
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):

    if plugin_config.subscription_bases:
        for base in plugin_config.subscription_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in plugin_config.subscription_bases
        ]
    else:
        for base in config.object_bases:
            registry.register_import(base)

            return [
                ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                for base in config.object_bases
            ]


def generate_operation(
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):

    tree = []

    if o.operation == OperationType.MUTATION:
        class_name = registry.generate_mutation_classname(o.name.value)
        extra_bases = get_mutation_bases(config, plugin_config, registry)
    if o.operation == OperationType.SUBSCRIPTION:
        class_name = registry.generate_subscription_classname(o.name.value)
        extra_bases = get_subscription_bases(config, plugin_config, registry)
    if o.operation == OperationType.QUERY:
        class_name = registry.generate_query_classname(o.name.value)
        extra_bases = get_query_bases(config, plugin_config, registry)

    x = get_operation_root_type(client_schema, o)
    operation_fields = []

    for field_node in o.selection_set.selections:

        field_node: FieldNode = field_node
        field_definition = get_field_def(client_schema, x, field_node)
        assert field_definition, "Couldn't find field definition"

        operation_fields += [
            type_field_node(
                field_node,
                field_definition,
                client_schema,
                config,
                tree,
                registry,
                parent_name=class_name,
            ),
        ]

    query_document = language.print_ast(o)
    z = fragment_searcher.findall(query_document)

    merged_document = replace_iteratively(query_document, registry)

    operation_fields += [
        ast.ClassDef(
            "Meta",
            bases=[],
            decorator_list=[],
            keywords=[],
            body=[
                ast.Assign(
                    targets=[ast.Name(id="domain", ctx=ast.Store())],
                    value=ast.Constant(value=str(config.domain)),
                ),
                ast.Assign(
                    targets=[ast.Name(id="document", ctx=ast.Store())],
                    value=ast.Constant(value=merged_document),
                ),
            ],
        )
    ]

    tree.append(
        ast.ClassDef(
            class_name,
            bases=extra_bases,
            decorator_list=[],
            keywords=[],
            body=operation_fields + generate_config_class(config),
        )
    )

    return tree


class OperationsPlugin(Plugin):
    def __init__(self, config=None, **data):
        self.plugin_config = config or OperationsPluginConfig(**data)

    def generate_ast(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        registry: ClassRegistry,
    ) -> List[ast.AST]:

        plugin_tree = []

        try:
            documents = parse_documents(
                client_schema, self.plugin_config.operations_glob or config.documents
            )
        except NoDocumentsFoundError as e:
            logger.exception(e)
            return plugin_tree

        definitions = documents.definitions
        operations = [
            node for node in definitions if isinstance(node, OperationDefinitionNode)
        ]

        for operation in operations:
            plugin_tree += generate_operation(
                operation, client_schema, config, self.plugin_config, registry
            )

        return plugin_tree



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['turms',
 'turms.cli',
 'turms.compat',
 'turms.parser',
 'turms.plugins',
 'turms.processor',
 'turms.stylers',
 'turms.types']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.0',
 'graphql-core>=3.2.0,<4.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'rich>=11.0.0,<12.0.0']

extras_require = \
{':python_version >= "3.7" and python_version < "3.9"': ['astunparse>=1.6.3,<2.0.0'],
 'black': ['black>=22.1.0,<23.0.0'],
 'isort': ['isort>=5.10.1,<6.0.0'],
 'watch': ['watchdog>=2.1.6,<3.0.0']}

entry_points = \
{'console_scripts': ['turms = turms.cli.main:entrypoint']}

setup_kwargs = {
    'name': 'turms',
    'version': '0.1.56',
    'description': 'graphql-codegen powered by pydantic',
    'long_description': '# turms\n\n[![codecov](https://codecov.io/gh/jhnnsrs/turms/branch/master/graph/badge.svg?token=UGXEA2THBV)](https://codecov.io/gh/jhnnsrs/turms)\n[![PyPI version](https://badge.fury.io/py/turms.svg)](https://pypi.org/project/turms/)\n[![Maintenance](https://img.shields.io/badge/Maintained%3F-yes-green.svg)](https://pypi.org/project/turms/)\n![Maintainer](https://img.shields.io/badge/maintainer-jhnnsrs-blue)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/turms.svg)](https://pypi.python.org/pypi/turms/)\n[![PyPI status](https://img.shields.io/pypi/status/turms.svg)](https://pypi.python.org/pypi/turms/)\n[![PyPI download month](https://img.shields.io/pypi/dm/turms.svg)](https://pypi.python.org/pypi/turms/)\n\n### DEVELOPMENT\n\n## Inspiration\n\nTurms is a pure python implementation of the awesome graphql-codegen library, following a simliar extensible design.\nIt makes heavy use of pydantic and its serialization capablities and provides fully typed querys, mutations and subscriptions\n\n## Supports\n\n- Documents\n- Fragments\n- Enums\n- Operations\n- Operation Functions\n- Scalar (mapping to python equivalent)\n\n## Features\n\n- Fully Modular (agnostic of graphql transport)\n- Tries to minimise Class Generation if using Fragments\n- Autocollapsing operation (if mutation or query has only one operation) functions\n- Specify type mixins, baseclasses...\n- Fully Support type hints for variables (Pylance)\n- Compliant with graphl-config\n\n## Companion Library\n\nIf you are searching for an Apollo-like GraphQL Client you can check out [rath](https://github.com/jhnnsrs/rath), that works especially\nwell with turms.\n\n## Installation\n\n```bash\npip install turms\n```\n\n## Configuration\n\nTurms configuration is compliant with the graphql-config specification, allowing interoperability with other frameworks.\n\n```yaml\nprojects:\n  default:\n    schema: http://api.spacex.land/graphql/\n    documents: graphql/**.graphql\n    extensions:\n      turms: # path for configuration for turms\n        out_dir: examples/api\n        stylers:\n          - type: turms.stylers.capitalize.Capitalizer\n          - type: turms.stylers.snake.SnakeNodeName\n        plugins:\n          - type: turms.plugins.enums.EnumsPlugin\n          - type: turms.plugins.inputs.InputsPlugin\n          - type: turms.plugins.fragments.FragmentsPlugin\n          - type: turms.plugins.operation.OperationsPlugin\n          - type: turms.plugins.funcs.OperationsFuncPlugin\n        processors:\n          - type: turms.processor.black.BlackProcessor\n          - type: turms.processor.isort.IsortProcessor\n        scalar_definitions:\n          uuid: str\n          timestamptz: str\n          Date: str\n```\n\nTurms configuration is based on plugins that can be configured in the graphql.config. There exist three major classes:\n\n### Stylers\n\nStylers are responsible for applying different styles to the class names and field names (e.g. snakecasing of graphqls pascalcase),\nthey are chained in order of notation in the graphql config (they receive the output of the last styler). Turms takes care of automatically\naliasing these names if they are not the same as the graphql style)\n\n### Plugins\n\nPlugins are the generators of code, that traverse through the direcotry and ad new ast nodes to the global tree. Examplary pluings are:\n\n- turms.plugins.enums.EnumsPlugin\n- turms.plugins.inputs.InputsPlugin\n- turms.plugins.fragments.FragmentsPlugin\n- turms.plugins.operation.OperationsPlugin\n- turms.plugins.funcs.OperationsFuncPlugin\n\n## Processors\n\nProcessors take the generated code (already a string), and can parse this code. (e.g. black processor for enforcing black style formatting) or isort of sorting imports.\nIncludes Processors are\n\n- turms.processor.black.BlackProcessor\n- turms.processor.isort.IsortProcessor\n\n## Usage\n\nOpen your workspace (create a virtual env), in the root folder\n\n```bash\nturms init\n```\n\nThis creates a graphql-config compliant configuration file in the working directory, edit this to reflect your settings (see Configuration)\n\n```bash\nturms gen\n```\n\nGenerate beautifully typed Operations, Enums,...\n\n### Why Turms\n\nIn Etruscan religion, Turms (usually written as 𐌕𐌖𐌓𐌌𐌑 Turmś in the Etruscan alphabet) was the equivalent of Roman Mercury and Greek Hermes, both gods of trade and the **messenger** god between people and gods.\n\n## Transport Layer\n\nTurms does not come with a default transport layer, but by specifiyng custom queries classes you can easily incorporate your logic (look at turms.types.herre for inspiration)\n\n## Examples\n\nThis github repository also contains an example graphql.config.yaml with the public SpaceX api, as well as a sample of the generated api.\n\n## Experimental\n\n```bash\nturms watch $PROJECT_NAME\n```\n\nTurms watch is able to automatically monitor your graphql folder for changes and autogenerate the api on save again.\nRequires additional dependency for watchdog\n\n```bash\npip install turms[watch]\n```\n',
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from graphql import (
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLScalarType,
    GraphQLWrappingType,
)
from turms.plugins.base import Plugin, PluginConfig
import ast
from typing import List, Optional
from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from turms.plugins.base import Plugin
from pydantic import BaseModel, BaseSettings
from graphql.type.definition import (
    GraphQLEnumType,
)
import keyword
from turms.registry import ClassRegistry


class InputsPluginConfig(PluginConfig):
    inputtype_bases: List[str] = ["pydantic.BaseModel"]
    skip_underscore: bool = True

    class Config:
        env_prefix = "TURMS_PLUGINS_INPUTS_"


def generate_input_annotation(
    type: GraphQLInputType,
    config: GeneratorConfig,
    plugin_config: InputsPluginConfig,
    registry: ClassRegistry,
    is_optional=True,
):
    if isinstance(type, GraphQLScalarType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Name(
                    id=registry.get_scalar_equivalent(type.name), ctx=ast.Load()
                ),
            )
        else:
            return ast.Name(
                id=registry.get_scalar_equivalent(type.name),
                ctx=ast.Load(),
            )

    if isinstance(type, GraphQLInputObjectType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Constant(value=registry.get_inputtype_class(type.name)),
            )
        return ast.Constant(
            value=registry.get_inputtype_class(type.name),
        )

    if isinstance(type, GraphQLEnumType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Constant(
                    value=registry.get_enum_class(type.name),
                ),
            )
        return ast.Constant(
            value=registry.get_enum_class(type.name),
        )
    if isinstance(type, GraphQLNonNull):
        return generate_input_annotation(
            type.of_type, config, plugin_config, registry, is_optional=False
        )

    if isinstance(type, GraphQLList):
        if is_optional:
            registry.register_import("typing.Optional")
            registry.register_import("typing.List")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Subscript(
                    value=ast.Name("List", ctx=ast.Load()),
                    slice=generate_input_annotation(
                        type.of_type, config, plugin_config, registry, is_optional=True
                    ),
                    ctx=ast.Load(),
                ),
            )

        registry.register_import("typing.List")
        return ast.Subscript(
            value=ast.Name("List", ctx=ast.Load()),
            slice=generate_input_annotation(
                type.of_type, config, plugin_config, registry, is_optional=True
            ),
            ctx=ast.Load(),
        )

    raise NotImplementedError(f"Unknown input type {type}")


def generate_inputs(
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: InputsPluginConfig,
    registry: ClassRegistry,
):

    tree = []

    inputobjects_type = {
        key: value
        for key, value in client_schema.type_map.items()
        if isinstance(value, GraphQLInputObjectType)
    }

    for base in plugin_config.inputtype_bases:
        registry.register_import(base)

    self_referential = set()

    for key, type in inputobjects_type.items():

        if plugin_config.skip_underscore and key.startswith("_"):
            continue

        name = registry.generate_inputtype_classname(key)
        fields = (
            [ast.Expr(value=ast.Constant(value=type.description))]
            if type.description
            else []
        )

        for value_key, value in type.fields.items():

            if isinstance(value.type, GraphQLInputObjectType):
                self_referential.add(name)

            field_name = registry.generate_node_name(value_key)

            if field_name != value_key:
                registry.register_import("pydantic.Field")
                assign = ast.AnnAssign(
                    target=ast.Name(field_name, ctx=ast.Store()),
                    annotation=generate_input_annotation(
                        value.type, config, plugin_config, registry, is_optional=True
                    ),
                    value=ast.Call(
                        func=ast.Name(id="Field", ctx=ast.Load()),
                        args=[],
                        keywords=[
                            ast.keyword(
                                arg="alias", value=ast.Constant(value=value_key)
                            )
                        ],
                    ),
                    simple=1,
                )
            else:
                assign = ast.AnnAssign(
                    target=ast.Name(value_key, ctx=ast.Store()),
                    annotation=generate_input_annotation(
                        value.type, config, plugin_config, registry, is_optional=True
                    ),
                    simple=1,
                )

            potential_comment = (
                value.description
                if not value.deprecation_reason
                else f"DEPRECATED: {value.description}"
            )

            if potential_comment:
                fields += [
                    assign,
                    ast.Expr(value=ast.Constant(value=potential_comment)),
                ]

            else:
                fields += [assign]

        registry.register_inputtype_class(key, name)

        tree.append(
            ast.ClassDef(
                name,
                bases=[
                    ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                    for base in plugin_config.inputtype_bases
                ],
                decorator_list=[],
                keywords=[],
                body=fields,
            )
        )

    for input_name in self_referential:
        tree.append(
            ast.Expr(
                value=ast.Call(
                    func=ast.Attribute(
                        value=ast.Name(
                            id=input_name,
                            ctx=ast.Load(),
                        ),
                        attr="update_forward_refs",
                        ctx=ast.Load(),
                    ),
                    keywords=[],
                    args=[],
                )
            )
        )

    return tree


class InputsPlugin(Plugin):
    def __init__(self, config=None, **data):
        self.plugin_config = config or InputsPluginConfig(**data)

    def generate_ast(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        registry: ClassRegistry,
    ) -> List[ast.AST]:

        for base in self.plugin_config.inputtype_bases:
            registry.register_import(base)

        return generate_inputs(client_schema, config, self.plugin_config, registry)



import json
from importlib import import_module
from typing import Optional
from urllib import request
import glob
import graphql
from turms.errors import GenerationError

from graphql import (
    build_ast_schema,
    build_client_schema,
    get_introspection_query,
    parse,
)


def import_class(module_path, class_name):
    module = import_module(module_path)
    return getattr(module, class_name)


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed. Simliar to
    djangos import_string, but without the cache.
    """

    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError(f"{dotted_path} doesn't look like a module path") from err

    try:
        return import_class(module_path, class_name)
    except AttributeError as err:
        raise ImportError(
            f"{module_path} does not define a {class_name} attribute/class"
        ) from err


def build_schema_from_introspect_url(
    schema_url: str, bearer_token: Optional[str] = None
) -> graphql.GraphQLSchema:
    """Introspect a GraphQL schema using introspection query

    Args:
        schema_url (_type_): The Schema url
        bearer_token (_type_, optional): A Bearer token. Defaults to None.

    Raises:
        GenerationError: _description_

    Returns:
        _type_: _description_
    """
    jdata = json.dumps({"query": get_introspection_query()}).encode("utf-8")
    req = request.Request(schema_url, data=jdata)
    req.add_header("Content-Type", "application/json")
    req.add_header("Accept", "application/json")
    if bearer_token:
        req.add_header("Authorization", f"Bearer {bearer_token}")

    try:
        resp = request.urlopen(req)
        x = json.loads(resp.read().decode("utf-8"))
    except Exception as e:
        raise GenerationError(f"Failed to fetch schema from {schema_url}")

    if "errors" in x:
        raise GenerationError(
            f"Failed to fetch schema from {schema_url} Graphql error: {x['errors']}"
        )

    return build_client_schema(x["data"])


def build_schema_from_glob(glob_string: str):
    schema_glob = glob.glob(glob_string, recursive=True)
    dsl_string = ""
    introspection_string = ""
    for file in schema_glob:
        with open(file, "r") as f:
            if file.endswith(".graphql"):
                dsl_string += f.read()
            elif file.endswith(".json"):
                # not really necessary as json files are generally not splitable
                introspection_string += f.read

    if not dsl_string and not introspection_string:
        raise GenerationError(f"No schema files found in {glob_string}")

    if dsl_string != "" and introspection_string != "":
        raise GenerationError("We cannot have both dsl and introspection files")
    if dsl_string != "":
        return build_ast_schema(parse(dsl_string))
    else:
        return build_client_schema(json.loads(introspection_string))



import argparse
from enum import Enum
import os
from turms.run import gen
from rich import get_console


directory = os.getcwd()


class TurmsOptions(str, Enum):
    GEN = "gen"
    INIT = "init"
    DOWNLOAD = "download"
    WATCH = "watch"


default_settings = """
projects:
  default:
    schema: https://api.spacex.land/graphql/
    documents: graphql/**.graphql
    extensions:
      turms:
        out_dir: examples/api
        stylers:
          - type: turms.styler.capitalize.Capitalizer
          - type: turms.styler.appender.Appender
        plugins:
          - type: turms.plugins.enums.EnumsPlugin
          - type: turms.plugins.inputs.InputsPlugin
          - type: turms.plugins.fragments.FragmentsPlugin
          - type: turms.plugins.operation.OperationsPlugin
          - type: turms.plugins.funcs.OperationsFuncPlugin
        processors:
          - type: turms.processor.black.BlackProcessor
        scalar_definitions:
          uuid: str
"""


def main(script: TurmsOptions, project=None):

    app_directory = os.getcwd()

    if script == TurmsOptions.INIT:
        get_console().print(f"Creating graphql.config.yaml in {app_directory}")
        with open(os.path.join(app_directory, "graphql.config.yaml"), "w") as f:
            f.write(default_settings)

    if script == TurmsOptions.WATCH:
        from turms.cli.watch import watch

        watch(os.path.join(app_directory, "graphql.config.yaml"), project=project)

    if script == TurmsOptions.GEN:
        gen(os.path.join(app_directory, "graphql.config.yaml"))


def entrypoint():
    parser = argparse.ArgumentParser(description="Say hello")
    parser.add_argument("script", type=TurmsOptions, help="The Script Type")
    parser.add_argument("project", type=str, help="The Path", nargs="?", default=None)
    args = parser.parse_args()
    try:
        main(script=args.script, project=args.project)
    except Exception as e:
        get_console().print_exception()


if __name__ == "__main__":
    entrypoint()
